import re

from .helpers import transform_encoding


class MessageTransformer:
    """parsing and transforming Slack messages"""

    def __init__(self, slack_service, locale_helper, font_family_mono_default) -> None:
        self._slack_service = slack_service
        self._locale_helper = locale_helper
        self._font_family_mono_default = font_family_mono_default

    def transform_text(self, text, use_mrkdwn=False):
        """transforms mrkdwn text into HTML text for PDF output

        Main method to resolve all mrkdwn, e.g. <C12345678>, <!here>, *bold*
        Will resolve channel and user IDs to their names if possible
        Returns string with rudimentary HTML for formatting and links

        Attr:
            text: text string to be transformed
            use_mrkdwn: will transform mrkdwn if set to true

        Returns:
            transformed text string with HTML formatting
        """

        def replace_mrkdwn_in_text(matchObj):
            """inline function returns replacement string for re.sub

            This function does the actual resolving of IDs and mrkdwn key words
            """
            match = matchObj.group(1)

            id_chars = match[0:2]
            id_raw = match[1 : len(match)]
            parts = id_raw.split("|", 1)
            obj_id = parts[0]

            make_bold = True
            if id_chars == "@U" or id_chars == "@W":
                # match is a user ID
                if obj_id in self._slack_service.user_names():
                    replacement = "@" + self._slack_service.user_names()[obj_id]
                else:
                    replacement = f"@user_{obj_id}"

            elif id_chars == "#C":
                # match is a channel ID
                if obj_id in self._slack_service.channel_names():
                    replacement = "#" + self._slack_service.channel_names()[obj_id]
                else:
                    replacement = f"#channel_{obj_id}"

            elif match[0:9] == "!subteam^":
                # match is a user group ID
                match2 = re.match(r"!subteam\^(S[A-Z0-9]+)", match)
                if match2 is not None and len(match2.groups()) == 1:
                    usergroup_id = match2.group(1)
                    if usergroup_id in self._slack_service.usergroup_names():
                        usergroup_name = self._slack_service.usergroup_names()[
                            usergroup_id
                        ]
                    else:
                        usergroup_name = f"usergroup_{usergroup_id}"
                else:
                    usergroup_name = "usergroup_unknown"
                replacement = "@" + usergroup_name

            elif match[0:1] == "!":
                # match is a special mention
                if obj_id == "here":
                    replacement = "@here"

                elif obj_id == "channel":
                    replacement = "@channel"

                elif obj_id == "everyone":
                    replacement = "@everyone"

                elif match[0:5] == "!date":
                    make_bold = False
                    date_parts = match.split("^")
                    if len(date_parts) > 1:
                        replacement = self._locale_helper.get_datetime_formatted_str(
                            date_parts[1]
                        )
                    else:
                        replacement = "(failed to parse date)"

                else:
                    replacement = f"@special_{obj_id}"

            else:
                # match is an URL
                link_parts = match.split("|")
                if len(link_parts) == 2:
                    url = link_parts[0]
                    text = link_parts[1]
                else:
                    url = match
                    text = match

                make_bold = False
                replacement = f'<a href="{url}">{text}</a>'

            if make_bold:
                replacement = f"<b>{replacement}</b>"

            return replacement

        # pass 1 - adjust encoding to latin-1 and transform HTML entities
        s2 = transform_encoding(text)

        # if requested try to transform mrkdwn in text
        if use_mrkdwn:

            # pass 2 - transform mrkdwns with brackets
            s2 = re.sub(r"<(.*?)>", replace_mrkdwn_in_text, s2)

            # pass 3 - transform formatting mrkdwns

            # bold
            s2 = re.sub(r"\*(.+)\*", r"<b>\1</b>", s2)

            # italic
            s2 = re.sub(r"\b_(.+)_\b", r"<i>\1</i>", s2)

            # code
            s2 = re.sub(
                r"`(.*)`",
                r'<s fontfamily="' + self._font_family_mono_default + r'">\1</s>',
                s2,
            )

            # indents
            s2 = re.sub(r"^>(.+)", r"<blockquote>\1</blockquote>", s2, 0, re.MULTILINE)

            s2 = s2.replace("</blockquote><br>", "</blockquote>")

            # EOF
            s2 = s2.replace("\n", "<br>")

        return s2



import datetime as dt
import html
import json
import os

from babel import Locale, UnknownLocaleError
from babel.dates import format_datetime, format_time, format_date
import pytz
from tzlocal import get_localzone


def transform_encoding(text):
    """adjust encoding to latin-1 and transform HTML entities"""
    text2 = html.unescape(text)
    text2 = text2.encode("utf-8", "replace").decode("utf-8")
    text2 = text2.replace("\t", "    ")
    return text2


def read_array_from_json_file(filename, quiet=False):
    """reads a json file and returns its contents as array"""
    filename += ".json"
    if not os.path.isfile(filename):
        if quiet is False:
            print(f"WARN: file does not exist: {filename}")
        arr = list()
    else:
        try:
            with open(filename, "r", encoding="utf-8") as f:
                arr = json.load(f)
        except IOError as e:
            if quiet is False:
                print(f"WARN: failed to read from {filename}: ", e)
            arr = list()

    return arr


def write_array_to_json_file(arr, filename):
    """writes array to a json file"""
    filename += ".json"
    print(f"Writing file: name {filename}")
    try:
        with open(filename, "w", encoding="utf-8") as f:
            json.dump(arr, f, sort_keys=True, indent=4, ensure_ascii=False)
    except IOError as e:
        print(f"ERROR: failed to write to {filename}: ", e)


class LocaleHelper:
    """Helpers for converting date & time according to current locale and timezone"""

    def __init__(
        self,
        my_locale: Locale = None,
        my_tz: pytz.BaseTzInfo = None,
        author_info: dict = None,
    ) -> None:
        """
        Args:
        - my_locale: Primary locale to use
        - my_tz: Primary timezone to use
        - author_info: locale and timezone to use from this Slack response
        if my_locale and/or my_tz are not given
        """
        # get locale to use
        if my_locale:
            if not isinstance(my_locale, Locale):
                raise TypeError("my_locale must be a babel Locale object")
        else:
            if author_info:
                try:
                    my_locale = Locale.parse(author_info["locale"], sep="-")
                except UnknownLocaleError:
                    print("WARN: Could not use locale info from Slack")
                    my_locale = Locale.default()
            else:
                my_locale = Locale.default()

        print(f"Locale is: {my_locale.get_display_name()} [{my_locale}]")
        self._locale = my_locale

        # get timezone to use
        if my_tz:
            if not isinstance(my_tz, pytz.BaseTzInfo):
                raise TypeError("my_tz must be of type pytz")
        else:
            if author_info:
                try:
                    my_tz = pytz.timezone(author_info["tz"])
                except pytz.exceptions.UnknownTimeZoneError:
                    print("WARN: Could not use timezone info from Slack")
                    my_tz = get_localzone()
            else:
                my_tz = get_localzone()

        print(f"Timezone is: {str(my_tz)}")
        self._timezone = my_tz

    @property
    def locale(self):
        return self._locale

    @property
    def timezone(self):
        return self._timezone

    def format_date_full_str(self, my_datetime):
        return format_date(my_datetime, format="full", locale=self.locale)

    def format_datetime_str(self, my_datetime):
        """returns formated datetime string for given dt using locale"""
        return format_datetime(my_datetime, format="short", locale=self.locale)

    def get_datetime_formatted_str(self, ts):
        """return given timestamp as formated datetime string using locale"""
        my_datetime = self.get_datetime_from_ts(ts)
        return format_datetime(my_datetime, format="short", locale=self.locale)

    def get_time_formatted_str(self, ts):
        """return given timestamp as formated datetime string using locale"""
        my_datetime = self.get_datetime_from_ts(ts)
        return format_time(my_datetime, format="short", locale=self.locale)

    def get_datetime_from_ts(self, ts):
        """returns datetime object of a unix timestamp with local timezone"""
        my_datetime = dt.datetime.fromtimestamp(float(ts), pytz.UTC)
        return my_datetime.astimezone(self.timezone)



# style and layout settings for PDF
PAGE_ORIENTATION_DEFAULT = "portrait"
PAGE_FORMAT_DEFAULT = "a4"
PAGE_UNITS_DEFAULT = "mm"
FONT_FAMILY_DEFAULT = "NotoSans"
FONT_FAMILY_MONO_DEFAULT = "NotoSansMono"
FONT_SIZE_NORMAL = 12
FONT_SIZE_LARGE = 14
FONT_SIZE_SMALL = 10
LINE_HEIGHT_DEFAULT = 6
LINE_HEIGHT_SMALL = 2
MARGIN_LEFT = 10
TAB_WIDTH = 4

# slack
MINUTES_UNTIL_USERNAME_REPEATS = 10
MAX_MESSAGES_PER_CHANNEL = 10000
SLACK_PAGE_LIMIT = 200
